from enum import Enum
from typing import List, Optional

from pydantic import BaseModel


class LoadedDataset(BaseModel):
    datasetId: str = None
    datasetNumOfRecords: int = None
    _hammingK: str = None
    _maxNQueries: str = None
    _neuralMatrixId: str = None
    _normalize: bool = None
    _rerankTopK: int = None
    searchType: str = None
    _typicalNQueries: int = None
    _inMemNumOfRecords: int
    _isLoaded: bool
    _pendingTransactionsInd: bool
    _removedIndexes: List
    _shiftMap: str


class EncodingDetails(BaseModel):
    binDatasetSizeInBytes: str = None
    binFilePath: str = None
    isActive: str = None
    nbits: str = None
    id: str = None


class CreateDatasetResponse(BaseModel):
    datasetId: str = None


class Document(BaseModel):
    document_id: str = None
    vector: List[float] = None
    metadata: Optional[List[dict]] = None


class DatasetStatus(str, Enum):
    completed = 'completed'
    training = 'training'
    error = 'error'
    loaded = 'loaded'


class TrainStatus(BaseModel):
    datasetStatus: DatasetStatus


class Response(BaseModel):
    status: str = None


class ResponseDeleteDocument(BaseModel):
    removeDataIndexes: List = None


class AddChunkRequest(BaseModel):
    documents: List[Document]


class Dataset(BaseModel):
    _datasetCopyFilePath: str = None
    _datasetFileType: str = None
    datasetName: str = None
    datasetStatus: str = None
    datasetType: str = None
    id: str = None
    numOfRecords: int = None
    numOfFeatures: int = None
    _encodingDetails: List[EncodingDetails] = None


class SearchResponse(BaseModel):
    distance: List[List[float]] = None
    indices: List[List[int]] = None
    metadata: Optional[List[Optional[List[Optional[dict]]]]] = None
    search: float
    total: float


class LoadedDatasets(LoadedDataset):
    __loaded_datasets: List[LoadedDataset] = None

    def get_loaded_datasets_list(self):
        return self.__loaded_datasets



from searchium.client.func import (
    validate_allocation,
    get_loaded_dataset,
    get_datasets,
    delete_dataset,
    load_dataset,
    unload_dataset,
    search,
    create_dataset,
    add_chunk,
    train_dataset,
    train_status,
    get_dataset,
    delete_document
)
from searchium.client.init import init



import json

import requests
from pydantic.tools import parse_obj_as

from searchium.client.init import Config
from searchium.fvs.model import *


def validate_allocation() -> bool:
    """
    Validate user allocation on cloud platform
    :return: bool
    """
    link = f"{Config.link}/board/allocation/validate"
    body = {"allocationId": f"{Config.allocation_id}"}
    get = requests.post(link, json=body)
    if get.status_code == 200:
        return True
    else:
        return False


def get_loaded_dataset() -> List[LoadedDataset]:
    """
    Get list loaded datasets
    :return: List[LoadedDataset]
    """
    link = f"{Config.link}/board/allocation/list"
    get = requests.get(link, headers=Config.headers)
    if get.status_code == 200:
        loaded = get.json()['allocationsList'][f'{Config.allocation_id}']['loadedDatasets']
        obj_as = parse_obj_as(List[LoadedDataset], loaded)
        return obj_as
    else:
        raise Exception(get.json())


def get_datasets() -> List[Dataset]:
    """
    Get list imported datasets
    :return: List[Dataset]
    """
    link = f"{Config.link}/dataset/list"
    get = requests.get(link, headers=Config.headers)
    if get.status_code == 200:
        list_datasets = get.json()['datasetsList']
        obj_as = parse_obj_as(List[Dataset], list_datasets)
        return obj_as
    else:
        raise Exception(get.json())


def get_dataset(dataset_id: str) -> Dataset:
    datasets: List[Dataset] = get_datasets()
    for data in datasets:
        if data.id == dataset_id:
            return data
    raise Exception(f"dataset by id: {dataset_id} is not found")


def delete_dataset(dataset_id: str) -> bool:
    """
    Delete dataset by dataset id
    :param dataset_id:
    :return: bool
    """
    link = f"{Config.link}/dataset/remove/{dataset_id}"
    get = requests.delete(link, headers=Config.headers)
    if get.status_code == 200:
        return True
    else:
        raise Exception(get.json())


def load_dataset(dataset_id: str) -> bool:
    """
    load dataset by dataset id
    :param dataset_id: str
    :return: bool
    """
    link = f"{Config.link}/dataset/load"
    body = {"allocationId": f"{Config.allocation_id}", "datasetId": f"{dataset_id}", "normalize": True,
            "typicalNQueries": 1, "maxNQueries": 50, "hammingK": 3200}
    get = requests.post(link, json=body, headers=Config.headers)
    if get.status_code == 200:
        return True
    else:
        raise Exception(get.json())


def unload_dataset(dataset_id: str) -> bool:
    """
    unload dataset by dataset id
    :param dataset_id: str
    :return: bool
    """
    link = f"{Config.link}/dataset/unload"
    body = {"allocationId": f"{Config.allocation_id}", "datasetId": f"{dataset_id}"}
    get = requests.post(link, json=body, headers=Config.headers)
    if get.status_code == 200:
        return True
    else:
        raise Exception(get.json())


def search(dataset_id: str, query: List[List[float]], topk: int = 5) -> SearchResponse:
    """
    search query in dataset
    return result object SearchResponse
    where fields are:
    distance - distance
    indices - indices
    metadata - optional
    search - netto apu search time
    total - full search time FVS
    :return: search result
    """
    link = f"{Config.link}/dataset/search"
    body = {"allocationId": f"{Config.allocation_id}", "datasetId": f"{dataset_id}", "topk": topk,
            "queriesFilePath": query}
    get = requests.post(link, json=body, headers=Config.headers)
    if get.status_code == 200:
        obj_as = parse_obj_as(SearchResponse, get.json())
        #return obj_as.dict(exclude_none=True)
        return obj_as
    else:
        raise Exception(get.json())


def create_dataset(dataset_id: Optional[str] = None) -> CreateDatasetResponse:
    """
    :return: str dataset_id
    """
    link = f"{Config.link}/dataset/create"
    body = {"datasetId": dataset_id}
    get = requests.post(link, json=body, headers=Config.headers)
    if get.status_code == 200:
        obj_as = parse_obj_as(CreateDatasetResponse, get.json())
        return obj_as
    else:
        raise Exception(get.json())


def train_dataset(dataset_id: str) -> Response:
    link = f"{Config.link}/dataset/train"
    body = {"datasetId": dataset_id}
    get = requests.post(link, json=body, headers=Config.headers)
    if get.status_code == 200:
        obj_as = parse_obj_as(Response, get.json())
        return obj_as
    else:
        raise Exception(get.json())


def train_status(dataset_id: str) -> TrainStatus:
    link = f"{Config.link}/dataset/train/status/{dataset_id}"
    get = requests.get(link, headers=Config.headers)
    if get.status_code == 200:
        obj_as = parse_obj_as(TrainStatus, get.json())
        return obj_as
    else:
        raise Exception(get.json())


def add_chunk(dataset_id: str, list_documents: List[dict]) -> Response:
    """
    :param dataset_id: str
    :param list_documents: List[dict] where dict contains
    {"document_id": str, vector: List[float], metadata: dict}
    :return: status ok
    """
    link = f"{Config.link}/dataset/add/chunk"
    records = [doc["vector"] for doc in list_documents]
    metadata = [doc["metadata"] if "metadata" in doc and (doc["metadata"] is not None) else None for doc in list_documents]
    document_ids = [doc["document_id"] for doc in list_documents]
    body = {"datasetId": dataset_id, "records": records, "metadata": metadata, "docsIds": document_ids}
    get = requests.post(link, json=body, headers=Config.headers)
    if get.status_code == 200:
        response = get.json()
        if "error" in response:
            raise Exception(response)
        return parse_obj_as(Response, response)
    else:
        raise Exception(get.json())
    pass


def delete_document(dataset_id: str, document_id: List[str] = None, delete_all: bool = False) -> ResponseDeleteDocument:
    link = f"{Config.link}/dataset/remove/documents"
    body = {"datasetId": dataset_id, "docsIds": document_id, "deleteAll": delete_all}
    get = requests.post(link, json=body, headers=Config.headers)
    if get.status_code == 200:
        print(f" response: {get.json()}")
        obj_as = parse_obj_as(ResponseDeleteDocument, get.json())
        return obj_as
    else:
        raise Exception(get.json())


def _check_connection() -> bool:
    # Alive
    # /alive
    pass



class Config(object):
    allocation_id: str = None
    url_searchium: str = None
    port: int = None
    link: str = None
    headers: str = None

    def __init__(self, allocation_id: str, url_searchium: str, cloud: bool = True, port: int = 7760, apiv: str = 'v1.0') -> None:
        if Config.allocation_id is None and Config.url_searchium is None:
            Config.allocation_id = allocation_id
            Config.url_searchium = url_searchium
            Config.port = port
            Config.link = f"{Config.url_searchium}/v1.0" if cloud else f"{Config.url_searchium}:{Config.port}/{apiv}"
            Config.headers = {'Content-Type': 'application/json', 'allocationToken': f'{Config.allocation_id}'}


def init(allocation: str, url: str, cloud: bool = True, port: int = 7760, apiv: str = 'v1.0') -> None:
    """
    Init configuration client of FVS Searchium
    All these params as allocation & url you can receive
     on our cloud platform searchium.ai
    :param apiv:
    :param cloud:
    :param allocation: str = instance id uuid
    :param port: as default 7760
    :param url: str
    :return: None
    """
    Config(allocation, url, cloud, port, apiv)



import uuid

import requests
from pydantic.tools import parse_obj_as

from searchium.fvs.model import *


class ClientFVS(object):
    def __init__(self, allocation_id: str, url_searchium: str, port: int = 7760) -> None:
        self._allocation = allocation_id
        self._url_searchium = url_searchium
        self._port = port
        self._link = f"{self._url_searchium}:{self._port}/v1.0"
        self._headers = {'Content-Type': 'application/json', 'allocationToken': f'{self._allocation}'}

    def validate_allocation(self) -> bool:
        """
        Validate user allocation on cloud platform
        :return: bool
        """
        link = f"{self._link}/board/allocation/validate"
        body = {"allocationId": f"{self._allocation}"}
        get = requests.post(link, json=body)
        if get.status_code == 200:
            return True
        else:
            return False

    def get_loaded_dataset(self) -> List[LoadedDataset]:
        """
        Get list loaded datasets
        :return: List[LoadedDataset]
        """
        link = f"{self._link}/board/allocation/list"
        get = requests.get(link, headers=self._headers)
        if get.status_code == 200:
            loaded = get.json()['allocationsList'][f'{self._allocation}']['loadedDatasets']
            obj_as = parse_obj_as(List[LoadedDataset], loaded)
            return obj_as
        else:
            raise Exception(get.json())

    def get_datasets(self) -> List[Dataset]:
        """
        Get list imported datasets
        :return: List[Dataset]
        """
        link = f"{self._link}/dataset/list"
        get = requests.get(link, headers=self._headers)
        if get.status_code == 200:
            list_datasets = get.json()['datasetsList']
            obj_as = parse_obj_as(List[Dataset], list_datasets)
            return obj_as
        else:
            raise Exception(get.json())

    def delete_dataset(self, dataset_id: str) -> bool:
        """
        Delete dataset by dataset id
        :param dataset_id:
        :return: bool
        """
        link = f"{self._link}/dataset/remove/{dataset_id}"
        get = requests.delete(link, headers=self._headers)
        if get.status_code == 200:
            return True
        else:
            raise Exception(get.json())

    def load_dataset(self, dataset_id: str) -> bool:
        """
        load dataset by dataset id
        :param dataset_id: str
        :return: bool
        """
        link = f"{self._link}/dataset/load"
        body = {"allocationId": f"{self._allocation}", "datasetId": f"{dataset_id}", "normalize": True,
                "typicalNQueries": 1, "maxNQueries": 50, "hammingK": 3200}
        get = requests.post(link, json=body, headers=self._headers)
        if get.status_code == 200:
            return True
        else:
            raise Exception(get.json())

    def unload_dataset(self, dataset_id: str) -> bool:
        """
        unload dataset by dataset id
        :param dataset_id: str
        :return: bool
        """
        link = f"{self._link}/dataset/unload"
        body = {"allocationId": f"{self._allocation}", "datasetId": f"{dataset_id}"}
        get = requests.post(link, json=body, headers=self._headers)
        if get.status_code == 200:
            return True
        else:
            raise Exception(get.json())

    def search(self, dataset_id: str, query: List[List[float]], topk: int = 5) -> SearchResponse:
        """
        search query in dataset
        return result object SearchResponse
        where fields are:
        distance - distance
        indices - indices
        metadata - optional
        search - netto apu search time
        total - full search time FVS
        :return: search result
        """
        link = f"{self._link}/dataset/search"
        body = {"allocationId": f"{self._allocation}", "datasetId": f"{dataset_id}", "topk": topk,
                "queriesFilePath": query}
        get = requests.post(link, json=body, headers=self._headers)
        if get.status_code == 200:
            obj_as = parse_obj_as(SearchResponse, get.json())
            return obj_as.dict(exclude_none=True)
        else:
            raise Exception(get.json())

    def create_dataset(self, dataset_name: Optional[str] = None) -> str:
        """
        :return: str dataset_id
        """
        link = f"{self._link}/dataset/create"
        get = requests.post(link, headers=self._headers)
        if get.status_code == 200:
            obj_as = parse_obj_as(str, get.json())
            return obj_as
        else:
            raise Exception(get.json())

    def _train_dataset(self, dataset_id):
        # /dataset/train
        # {
        #     "datasetId": "",
        #     "trainType": "Regular"
        # }
        pass
    
    def _add_chunk(self, dataset_id: str) -> None:
        pass

    def _check_connection(self) -> bool:
        # Alive
        # /alive
        pass


def get_client(allocation: str, url: str, port: int = 7760) -> ClientFVS:
    """
    Get client of FVS Searchium
    All these params as allocation & url you can receive
     on our cloud platform searchium.ai
    :param port: as default 7760
    :param allocation: str uuid
    :param url: str
    :return: client of FVS ClientFVS
    """
    return ClientFVS(allocation, url, port)
